from enum import Enum

from handler_app import plh  # type: ignore


@plh.get("/teapot", status_code=418)
def hello_teapot():
    return {"message": "I'm a teapot"}


@plh.get("/pets/{petId}", operation_id="Create Pet")
def pets_handler(petId):
    return {"pet_id": petId}


@plh.get("/items/{item_id}")
def handler_with_type_hint(item_id: int):
    return {"item_id": item_id}


class Animals(str, Enum):
    dog = "dog"


@plh.get("/item_enum/{item_id}")
def handler_with_enum_type_hint(item_id: Animals):
    return {"item_id": item_id}



from typing import Optional


class Response:
    def __init__(self, content=None, status_code=None, headers: Optional[dict[str, str]] = None, media_type=None):
        """
        content - A str or bytes.
        status_code - An int HTTP status code.
        headers - A dict of strings.
        media_type - A str giving the media type. E.g. "text/html".
        """
        self.content = content
        self.status_code = status_code
        self.headers = headers
        self.media_type = media_type


class GpkgFileResponse(Response):
    def __init__(
        self,
        content,
        file_name=None,
    ):
        """
        content - A str or bytes.
        status_code - An int HTTP status code.
        headers - A dict of strings.
        media_type - A str giving the media type. E.g. "text/html".
        """
        super().__init__(
            content=content,
            headers={"Content-Type": "application/xml", "Content-Disposition": f'attachment; filename="{file_name}"'},
        )



from typing import Optional

from handler_app import plh
from pydantic import BaseModel


class FunModel(BaseModel):
    item_name: str
    item_value: Optional[int]


class ListFunModel(BaseModel):
    __root__: list[FunModel]


@plh.get("/response_model", response_model=FunModel)
def response_model(secret):
    return {"item_name": secret}


@plh.get("/list_response")
def response_list():
    return [{"item_name": 1}]


@plh.get("/list_response_model", response_model=ListFunModel)
def response_list_model():
    """List models"""
    return [{"item_name": "secret"}]



from datetime import datetime
from decimal import Decimal
from typing import Optional, Union

from handler_app import plh

from pydantic_lambda_handler.params import Query

fake_items_db = [{"item_name": "Foo"}, {"item_name": "Bar"}, {"item_name": "Baz"}]


@plh.get("/query")
def query_skip(skip: int = 0, limit: int = 10):
    return fake_items_db[skip : skip + limit]


@plh.get("/query_param")
def query_param(sausages: Optional[str] = Query(default=None, alias="meat")):
    return {"sausages": sausages}


@plh.get("/query_multivalue_param")
def query_multi_param(sausages: Optional[list[int]]):
    return {"sausages": sausages}


@plh.get("/query_required")
def query_required(secret):
    return {"item_name": secret}


@plh.get("/query_float")
def query_float(item_name: float):
    return {"item_name": item_name}


@plh.get("/query_union")
def query_union(
    param: Union[Decimal, datetime, None] = None,
):
    return {"param": param}



from typing import Any

from awslambdaric.lambda_context import LambdaContext

from pydantic_lambda_handler.models import BaseOutput


class BaseHook:
    @staticmethod
    def method_init(**kwargs) -> None:
        """1st"""
        return

    @staticmethod
    def post_create_response(**kwargs) -> None:
        """
        2nd: This is called after the hook is created.
        """
        return

    @staticmethod
    def pre_path(**kwargs) -> None:
        """3rd"""
        return

    @staticmethod
    def pre_func(event, context) -> tuple[dict, LambdaContext]:
        """4th"""
        return event, context

    @staticmethod
    def post_func(body) -> Any:
        """5th"""
        return body

    @staticmethod
    def pre_return(output: BaseOutput) -> BaseOutput:
        """6th"""
        return output



import json
import os
import re
from pathlib import Path

import pytest
import requests
from awslambdaric.lambda_context import LambdaContext

from pydantic_lambda_handler.gen_open_api_inspect import gen_open_api_inspect


def pytest_addoption(parser):
    parser.addoption("--live", action="store_true", help="Also test against real AWS")


class Response:
    def __init__(self, response):
        self._response = response

    @property
    def status_code(self):
        return self._response["statusCode"]

    def json(self):
        return json.loads(self._response["body"])

    @property
    def text(self):
        return self._response["body"]

    @property
    def content(self):
        return bytes(self._response["body"])


class RequestClient:
    def __init__(self):
        path = Path(__file__).parents[1].joinpath("demo_app/src")
        self._spec, self._cdk_stuff, self._test = gen_open_api_inspect(path)

    def get(self, url, *args, **kwargs):
        try:
            return self._mock_request(url, "get", *args, **kwargs)
        except Exception:
            return Response({"statusCode": 502, "body": json.dumps({"message": "Internal server error"})})

    def post(self, url, *args, **kwargs):
        try:
            return self._mock_request(url, "post", *args, **kwargs)
        except Exception:
            return Response({"statusCode": 502, "body": json.dumps({"message": "Internal server error"})})

    def _mock_request(self, url, method, *args, **kwargs):
        event = self.generate_event(kwargs, url)

        for comp_url, info in self._test["paths"].items():
            try:
                match = re.fullmatch(comp_url, url)
            except Exception as e:
                print(e)
                # None breaking here
                continue

            if match:
                decorated_function_ = info[method]["handler"]["decorated_function"]
                break
        else:  # No break
            raise ValueError

        context = LambdaContext(
            invoke_id="abd",
            client_context=None,
            cognito_identity=None,
            epoch_deadline_time_in_ms=1660605740936,
            invoked_function_arn="abd",
        )

        function_ = self._test["paths"][comp_url][method]["handler"]["function"]
        response = decorated_function_(function_)(event, context)
        return Response(response)

    def generate_event(self, kwargs, url):
        if "data" in kwargs:
            body = kwargs["data"]
        elif "json" in kwargs:
            body = json.dumps(kwargs["json"])
        else:
            body = None
        for comp_url, info in self._test["paths"].items():
            try:
                match = re.fullmatch(comp_url, url)
            except Exception:
                # None breaking here
                continue

            if match:
                break
        else:  # No break
            raise ValueError

        qs_params = {}
        mult_qs_params = {}

        for key, values in kwargs.get("params", {}).items():
            if isinstance(values, list):
                for value in values:
                    qs_params[key] = value
                mult_qs_params[key] = values
            else:
                qs_params[key] = values
                mult_qs_params[key] = [values]

        event = {
            "body": body,
            "queryStringParameters": qs_params or None,
            "multiValueQueryStringParameters": mult_qs_params or None,
            "pathParameters": match.groupdict(),
            "headers": kwargs.get("headers"),
        }
        return event


@pytest.fixture(scope="function")
def base_url(requests_client_type):
    return os.environ["BASE_URL"] if requests_client_type == "real" else ""


@pytest.fixture(scope="function")
def requests_client(requests_client_type):
    return requests if requests_client_type == "real" else RequestClient()


def pytest_generate_tests(metafunc):
    if "requests_client_type" in metafunc.fixturenames:
        types = ["mock"]
        if metafunc.config.getoption("live"):
            types.append("real")
        metafunc.parametrize("requests_client_type", types)


@pytest.fixture
def _gen():
    path = Path(__file__).parents[1].joinpath("demo_app/src")
    schema, cdk_stuff, _ = gen_open_api_inspect(path)
    return schema, cdk_stuff


@pytest.fixture
def schema(_gen):
    schema, *_ = _gen
    return json.loads(schema)


@pytest.fixture
def cdk_config(_gen):
    _, cdk_config = _gen
    return cdk_config


@pytest.fixture
def mock_lambda_context():
    return LambdaContext(
        invoke_id="abd",
        client_context=None,
        cognito_identity=None,
        epoch_deadline_time_in_ms=1660605740936,
        invoked_function_arn="abd",
    )



"""
https://docs.aws.amazon.com/lambda/latest/dg/services-apigateway-tutorial.html#services-apigateway-tutorial-prereqs
"""
import subprocess
import sys
from pathlib import Path
from pprint import pprint
from tempfile import mkdtemp

from aws_cdk import Stack
from aws_cdk import aws_apigateway as _apigw
from aws_cdk import aws_iam as iam
from aws_cdk import aws_lambda
from constructs import Construct

from pydantic_lambda_handler.gen_open_api_inspect import gen_open_api_inspect


def build_requirements():
    requirements_dir = Path(mkdtemp())
    root = Path(__file__).parents[2]

    # https://aws.amazon.com/premiumsupport/knowledge-center/lambda-python-package-compatible/
    # pip install \
    #     --platform manylinux2014_x86_64 \
    #     --target=my-lambda-function \
    #     --implementation cp \
    #     --python 3.9 \
    #     --only-binary=:all: --upgrade \
    #     pandas
    subprocess.run(
        (
            f"{sys.executable}",
            "-m",
            "pip",
            "install",
            f"{root}",
            "-r",
            f"{root.joinpath('demo_app/src/requirements.txt')}",
            "--upgrade",
            "--target",
            requirements_dir.joinpath("python"),
            "--platform",
            "manylinux2014_x86_64",
            "--implementation",
            "cp",
            "--python",
            "3.9",
            "--only-binary=:all:",
        ),
        check=True,
    )

    return requirements_dir


class DemoAppStack(Stack):
    def __init__(self, scope: Construct, construct_id: str, **kwargs) -> None:
        super().__init__(scope, construct_id, **kwargs)

        policy_statement_1 = iam.PolicyStatement(
            actions=[
                "dynamodb:DeleteItem",
                "dynamodb:GetItem",
                "dynamodb:PutItem",
                "dynamodb:Query",
                "dynamodb:Scan",
                "dynamodb:UpdateItem",
            ],
            effect=iam.Effect.ALLOW,
            resources=["*"],
        )

        policy_statement_2 = iam.PolicyStatement(
            actions=["logs:CreateLogGroup", "logs:CreateLogStream", "logs:PutLogEvents"],
            effect=iam.Effect.ALLOW,
            resources=["*"],
        )

        document = iam.PolicyDocument(statements=[policy_statement_1, policy_statement_2])

        iam.Policy(
            self,
            "id_123",
            policy_name="lambda-apigateway-policy",
            document=document,
        )

        user = iam.User(self, "myuser")

        iam.Role(self, "id_1234", role_name="lambda-apigateway-role", assumed_by=user)

        repo_dir = Path(__file__).parents[2]

        requirements_dir = build_requirements()

        base_api = _apigw.RestApi(self, "ApiGatewayWithCors", rest_api_name="ApiGatewayWithCors")

        base_lambda_layer = aws_lambda.LayerVersion(
            self,
            "base_layer",
            code=aws_lambda.Code.from_asset(str(requirements_dir)),
        )

        app_dir = repo_dir.joinpath("demo_app/src")
        _, cdk_config, _ = gen_open_api_inspect(app_dir)

        self.add_resources(app_dir, base_api.root, base_lambda_layer, cdk_config)

    def add_resources(self, app_dir, resource, base_lambda_layer, config):
        for conf in config:
            if conf.get("name"):
                new_resource = resource.add_resource(conf["name"])
            else:
                new_resource = resource

            for method_conf in conf.get("methods", ()):
                pprint(method_conf)
                self.add_method(method_conf, new_resource, app_dir, base_lambda_layer)

            self.add_resources(app_dir, new_resource, base_lambda_layer, conf.get("resources", ()))

    def add_method(self, method_conf, resource, app_dir, base_lambda_layer):
        pprint(method_conf)
        func = aws_lambda.Function(
            self,
            method_conf["function_name"],
            runtime=aws_lambda.Runtime.PYTHON_3_9,
            handler=method_conf["reference"],
            layers=[base_lambda_layer],
            code=aws_lambda.Code.from_asset(
                str(app_dir),
            ),
        )

        api_gw_lambda = _apigw.LambdaIntegration(
            func,
            proxy=True,
            integration_responses=[
                _apigw.IntegrationResponse(
                    status_code=method_conf["status_code"],
                    response_parameters={"method.response.header.Access-Control-Allow-Origin": "'*'"},
                )
            ],
        )

        resource.add_method(
            method_conf["method"],
            api_gw_lambda,
            method_responses=[
                _apigw.MethodResponse(
                    status_code=method_conf["status_code"],
                    response_parameters={"method.response.header.Access-Control-Allow-Origin": True},
                )
            ],
        )



from handler_app import plh


@plh.get("/error", errors=[(418, ValueError)])
def error_skip():
    raise ValueError("nope")


@plh.get("/multiple_errors", errors=[(422, (ValueError, TypeError))])
def error_multiple():
    raise ValueError("nope")



from collections import defaultdict
from pathlib import Path
from typing import Any

from awslambdaric.lambda_context import LambdaContext

from pydantic_lambda_handler.middleware import BaseHook


class CDKConf(BaseHook):
    """Gen cdk conf"""

    _dir_path: Path
    _method = None
    _hold_dict: dict = defaultdict(dict)

    @classmethod
    def method_init(cls, **kwargs):
        cls._method = kwargs["method"].upper()
        cls._hold_dict[kwargs["url"]].update(
            {
                cls._method: {
                    "function_name": kwargs["function_name"],
                    "status_code": str(int(kwargs["status_code"])),
                }
            }
        )

    @classmethod
    def pre_path(cls, **kwargs) -> None:
        func = kwargs["func"]
        cls._hold_dict[kwargs["url"]][cls._method]["index"] = str(f'{func.__module__.replace(".", "/")}.py')
        cls._hold_dict[kwargs["url"]][cls._method]["handler"] = func.__name__
        cls._hold_dict[kwargs["url"]][cls._method]["reference"] = f"{func.__module__}.{func.__qualname__}"
        cls._hold_dict[kwargs["url"]][cls._method]["function_name"] = cls._hold_dict[kwargs["url"]][cls._method][
            "function_name"
        ] or to_camel_case(func.__name__)

    @classmethod
    def pre_func(cls, event, context) -> tuple[dict, LambdaContext]:
        return event, context

    @classmethod
    def post_func(cls, body) -> Any:
        return body

    @classmethod
    def generate(cls):
        methods = ("GET", "POST")

        resource = []
        for url, conf in sorted(cls._hold_dict.items()):
            add_resource_v2(resource, url, conf)

        def sort_recursive(resource_list: list[dict]):
            resource_list.sort(key=lambda x: x.get("name", ""))
            seen = set()
            for i in resource_list:
                if "methods" in i:
                    i["methods"] = sorted(i["methods"], key=lambda x: methods.index(x["method"]))

                if "name" in i:
                    if i["name"] in seen:
                        raise ValueError(f'{i["name"]=}')

                    seen.add(i["name"])

                if "resources" in i:
                    sort_recursive(i["resources"])

        sort_recursive(resource)
        return resource


def add_resource_v2(child_list: list[dict], url: str, conf):
    name, found, remaining = url.partition("/")

    if not name and not remaining:
        # at "/"
        methods: list[dict] = []
        child_list.append({"methods": methods, "name": name})
        for method, method_conf in conf.items():
            methods.append(add_method_v2(method, method_conf))
        return
    elif not name and remaining:
        # "", "/" query
        if not child_list:
            child_resource = {}
            resources: list[dict] = []
            child_resource["resources"] = resources
            child_list.append(child_resource)
            add_resource_v2(resources, remaining, conf)
        else:
            child_list_ = child_list[0]
            if "resources" in child_list_:
                resources = child_list_["resources"]
            else:
                resources = []
                child_list_["resources"] = resources

            add_resource_v2(resources, remaining, conf)

    elif name and remaining:
        # need to keep traversing
        child_resource = next((i for i in child_list if i.get("name") == name), None)  # type: ignore
        if not child_resource:
            resources = []
            child_resource = {"name": name, "resources": resources}  # type: ignore
            child_list.append(child_resource)
            add_resource_v2(resources, remaining, conf)
        else:
            if "resources" in child_resource:
                resources = child_resource["resources"]
            else:
                resources = []
                child_resource["resources"] = resources

            add_resource_v2(resources, remaining, conf)
    else:
        # add methods
        child_resource = next((i for i in child_list if i.get("name") == name), None)  # type: ignore
        if not child_resource:
            child_resource = {"name": name}  # type: ignore
            child_list.append(child_resource)
        else:
            print()

        if "methods" in child_resource:
            methods = child_resource["methods"]
        else:
            methods = []
            child_resource["methods"] = methods

        for method, method_conf in conf.items():
            methods.append(add_method_v2(method, method_conf))

    return child_list


def to_camel_case(text):
    s = text.replace("-", " ").replace("_", " ")
    s = s.split()
    if len(text) == 0:
        return text.capitalize()
    return "".join(i.capitalize() for i in s)


def add_method_v2(method, conf):
    return {
        "method": method,
        "index": conf["index"],
        "handler": conf["handler"],
        "reference": conf["reference"],
        "status_code": conf["status_code"],
        "function_name": conf["function_name"],
    }



import json
import os
import sys
from ast import Import, ImportFrom, parse, walk
from copy import deepcopy
from importlib.util import module_from_spec, spec_from_file_location
from inspect import getmembers
from pathlib import Path
from sys import modules
from typing import Optional

from pydantic_lambda_handler.hooks.cdk_conf_hook import CDKConf
from pydantic_lambda_handler.hooks.mock_requests import MockRequests
from pydantic_lambda_handler.hooks.open_api_gen_hook import APIGenerationHook
from pydantic_lambda_handler.main import PydanticLambdaHandler


def get_top_imported_names(file: str) -> set[str]:
    """Collect names imported in given file.

    We only collect top-level names, i.e. `from foo.bar import baz`
    will only add `foo` to the list.
    """
    if not file.endswith(".pyi"):
        return set()
    with open(os.path.join(file), "rb") as f:
        content = f.read()
    parsed = parse(content)
    top_imported = set()
    for node in walk(parsed):
        if isinstance(node, Import):
            for name in node.names:
                top_imported.add(name.name.split(".")[0])
        elif isinstance(node, ImportFrom):
            if node.level > 0:
                # Relative imports always refer to the current package.
                continue
            assert node.module
            top_imported.add(node.module.split(".")[0])
    return top_imported


class add_path:
    def __init__(self, path):
        self.path = str(path)
        self.added = False

    def __enter__(self):
        if self.path not in sys.path:
            sys.path.insert(0, self.path)
            self.added = True

    def __exit__(self, exc_type, exc_value, traceback):
        if self.added:
            try:
                sys.path.remove(self.path)
            except ValueError:
                pass


def gen_open_api_inspect(dir_path: Path):
    with add_path(dir_path):
        files = list(dir_path.glob("**/*handler[s].py"))
        if t_files := list(dir_path.glob("handlers/**/*.py")):
            files.extend(t_files)

        app_fp = dir_path.joinpath("app.py")
        if app_fp.exists():
            files.insert(0, app_fp)
        else:
            files.insert(0, dir_path.joinpath("handler_app.py"))

        # This is ugly, once it works refactor out
        PydanticLambdaHandler.add_hook(APIGenerationHook)
        CDKConf._dir_path = dir_path
        PydanticLambdaHandler.add_hook(CDKConf)
        PydanticLambdaHandler.add_hook(MockRequests)

        app: Optional[PydanticLambdaHandler] = None

        for file in files:
            module_name = ".".join(str(file.relative_to(dir_path)).removesuffix(".py").split("/"))
            spec = spec_from_file_location(module_name, file)
            if not spec or not spec.loader:
                continue
            module = module_from_spec(spec)
            modules[module_name] = module
            spec.loader.exec_module(module)
            results = getmembers(module)

            for i in range(len(results)):
                if isinstance(results[i][1], PydanticLambdaHandler):
                    app = deepcopy(results[i][1])

        if app:
            return (
                next(h for h in app._hooks if issubclass(h, APIGenerationHook)).generate(),  # type: ignore
                next(h for h in app._hooks if issubclass(h, CDKConf)).generate(),  # type: ignore
                next(h for h in app._hooks if issubclass(h, MockRequests)).testing_stuff,  # type: ignore
            )
        raise ValueError("App not found")


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("app_dir")
    parser.add_argument("output_file_path")
    args = parser.parse_args()

    app_dir = args.app_dir
    output_file_path = args.output_file_path

    app_dir_path = Path(app_dir)
    app_dir_path.exists()

    schema, *_ = gen_open_api_inspect(app_dir_path)

    output_file_path = Path(output_file_path)
    output_file_path.parent.exists()

    schema = json.loads(schema)

    with output_file_path.open("w") as f:
        json.dump(schema, f, indent=4)


def cdk_conf():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("app_dir")
    parser.add_argument("output_file_path")
    args = parser.parse_args()

    app_dir = args.app_dir
    output_file_path = args.output_file_path

    app_dir_path = Path(app_dir)
    app_dir_path.exists()

    _, cdk_conf, _ = gen_open_api_inspect(app_dir_path)

    output_file_path = Path(output_file_path)
    output_file_path.parent.exists()

    with output_file_path.open("w") as f:
        json.dump(cdk_conf, f, indent=4)



from pydantic_lambda_handler.hooks.cdk_conf_hook import CDKConf, add_resource_v2


def test_generate_cdk_config(cdk_config):
    hello_resource = next(i for i in cdk_config[0]["resources"] if i.get("name") == "hello")
    assert hello_resource["name"] == "hello", cdk_config["resources"]
    assert hello_resource["methods"] == [
        {
            "function_name": "HelloHandler",
            "handler": "hello_handler",
            "index": "demo_app_handlers.py",
            "method": "GET",
            "reference": "demo_app_handlers.hello_handler",
            "status_code": "200",
        },
        {
            "function_name": "CreateHandler",
            "handler": "create_handler",
            "index": "demo_app_handlers.py",
            "method": "POST",
            "reference": "demo_app_handlers.create_handler",
            "status_code": "201",
        },
    ]


def test_generate_cdk_config_no_name(mocker):
    """
    Bug: if path "/" doesn't exist then a KeyError would occur
    """
    hold_dict = {
        "/query": {
            "GET": {
                "function_name": "QuerySkip",
                "status_code": "200",
                "index": "subfolder/query_parameters_handlers.py",
                "handler": "query_skip",
                "reference": "subfolder.query_parameters_handlers.query_skip",
            }
        },
    }
    mocker.patch.object(CDKConf, "_hold_dict", return_value=hold_dict)
    CDKConf.generate()


def test_generate_cdk_config_status_code(cdk_config):
    hello_resource = next(i for i in cdk_config[0]["resources"] if i.get("name") == "hello")
    assert hello_resource["methods"][0]["status_code"] == "200"


def test_generate_cdk_config_nested_resources(cdk_config):
    hello_resource = next(i for i in cdk_config[0]["resources"] if i.get("name") == "items")
    assert hello_resource["resources"] == [
        {
            "methods": [
                {
                    "function_name": "HandlerWithTypeHint",
                    "handler": "handler_with_type_hint",
                    "index": "subfolder/path_parameters_handlers.py",
                    "method": "GET",
                    "reference": "subfolder.path_parameters_handlers.handler_with_type_hint",
                    "status_code": "200",
                }
            ],
            "name": "{item_id}",
        }
    ]


conf = [
    (
        "/hello",
        {
            "POST": {
                "function_name": "CreateHandler",
                "status_code": "201",
                "index": "demo_app_handlers.py",
                "handler": "create_handler",
                "reference": "demo_app_handlers.create_handler",
            },
            "GET": {
                "function_name": "HelloHandler",
                "status_code": "200",
                "index": "demo_app_handlers.py",
                "handler": "hello_handler",
                "reference": "demo_app_handlers.hello_handler",
            },
        },
    ),
    (
        "/",
        {
            "POST": {
                "function_name": "IndexHandler",
                "status_code": "201",
                "index": "demo_app_handlers.py",
                "handler": "index_handler",
                "reference": "demo_app_handlers.index_handler",
            }
        },
    ),
    (
        "/query",
        {
            "GET": {
                "function_name": "QuerySkip",
                "status_code": "200",
                "index": "subfolder/query_parameters_handlers.py",
                "handler": "query_skip",
                "reference": "subfolder.query_parameters_handlers.query_skip",
            }
        },
    ),
    (
        "/query_required",
        {
            "GET": {
                "function_name": "QueryRequired",
                "status_code": "200",
                "index": "subfolder/query_parameters_handlers.py",
                "handler": "query_required",
                "reference": "subfolder.query_parameters_handlers.query_required",
            }
        },
    ),
    (
        "/teapot",
        {
            "GET": {
                "function_name": "HelloTeapot",
                "status_code": "418",
                "index": "subfolder/path_parameters_handlers.py",
                "handler": "hello_teapot",
                "reference": "subfolder.path_parameters_handlers.hello_teapot",
            }
        },
    ),
    (
        "/pets/{petId}",
        {
            "GET": {
                "function_name": "PetsHandler",
                "status_code": "200",
                "index": "subfolder/path_parameters_handlers.py",
                "handler": "pets_handler",
                "reference": "subfolder.path_parameters_handlers.pets_handler",
            }
        },
    ),
    (
        "/items/{item_id}",
        {
            "GET": {
                "function_name": "HandlerWithTypeHint",
                "status_code": "200",
                "index": "subfolder/path_parameters_handlers.py",
                "handler": "handler_with_type_hint",
                "reference": "subfolder.path_parameters_handlers.handler_with_type_hint",
            }
        },
    ),
    (
        "/item_enum/{item_id}",
        {
            "GET": {
                "function_name": "HandlerWithEnumTypeHint",
                "status_code": "200",
                "index": "subfolder/path_parameters_handlers.py",
                "handler": "handler_with_enum_type_hint",
                "reference": "subfolder.path_parameters_handlers.handler_with_enum_type_hint",
            }
        },
    ),
    (
        "/context",
        {
            "GET": {
                "function_name": "WithContext",
                "status_code": "200",
                "index": "subfolder/additional_args_handlers.py",
                "handler": "with_context",
                "reference": "subfolder.additional_args_handlers.with_context",
            }
        },
    ),
    (
        "/response_model",
        {
            "GET": {
                "function_name": "ResponseModel",
                "status_code": "200",
                "index": "subfolder/response_model_handlers.py",
                "handler": "response_model",
                "reference": "subfolder.response_model_handlers.response_model",
            }
        },
    ),
]


def test_add_resource_root_v2():
    child_list = []
    url, conf = (
        "/",
        {
            "POST": {
                "function_name": "IndexHandler",
                "status_code": "201",
                "index": "demo_app_handlers.py",
                "handler": "index_handler",
                "reference": "demo_app_handlers.index_handler",
            }
        },
    )
    add_resource_v2(child_list, url, conf)
    assert child_list == [
        {
            "name": "",
            "methods": [
                {
                    "function_name": "IndexHandler",
                    "handler": "index_handler",
                    "index": "demo_app_handlers.py",
                    "method": "POST",
                    "reference": "demo_app_handlers.index_handler",
                    "status_code": "201",
                }
            ],
        }
    ]


def test_add_resource_query_v2():
    child_list = []
    url, conf = (
        "/query",
        {
            "GET": {
                "function_name": "QuerySkip",
                "status_code": "200",
                "index": "subfolder/query_parameters_handlers.py",
                "handler": "query_skip",
                "reference": "subfolder.query_parameters_handlers.query_skip",
            }
        },
    )
    add_resource_v2(child_list, url, conf)
    assert child_list == [
        {
            "resources": [
                {
                    "name": "query",
                    "methods": [
                        {
                            "function_name": "QuerySkip",
                            "handler": "query_skip",
                            "index": "subfolder/query_parameters_handlers.py",
                            "method": "GET",
                            "reference": "subfolder.query_parameters_handlers.query_skip",
                            "status_code": "200",
                        }
                    ],
                }
            ]
        }
    ]


def test_add_resource_deep_v2():
    child_list = []
    url, conf = (
        "/pets/{petId}",
        {
            "GET": {
                "function_name": "PetsHandler",
                "status_code": "200",
                "index": "subfolder/path_parameters_handlers.py",
                "handler": "pets_handler",
                "reference": "subfolder.path_parameters_handlers.pets_handler",
            }
        },
    )

    add_resource_v2(child_list, url, conf)
    assert child_list == [
        {
            "resources": [
                {
                    "name": "pets",
                    "resources": [
                        {
                            "name": "{petId}",
                            "methods": [
                                {
                                    "function_name": "PetsHandler",
                                    "handler": "pets_handler",
                                    "index": "subfolder/path_parameters_handlers.py",
                                    "method": "GET",
                                    "reference": "subfolder.path_parameters_handlers.pets_handler",
                                    "status_code": "200",
                                }
                            ],
                        }
                    ],
                }
            ],
        }
    ]


def test_add_resource_deep_plus2_v2():
    child_list = []
    child_dict = {
        "/pets/{petId}": {
            "GET": {
                "function_name": "PetsHandler",
                "status_code": "200",
                "index": "subfolder/path_parameters_handlers.py",
                "handler": "pets_handler",
                "reference": "subfolder.path_parameters_handlers.pets_handler",
            }
        },
        "/context": {
            "GET": {
                "function_name": "WithContext",
                "status_code": "200",
                "index": "subfolder/additional_args_handlers.py",
                "handler": "with_context",
                "reference": "subfolder.additional_args_handlers.with_context",
            }
        },
    }

    for url, conf in child_dict.items():
        add_resource_v2(child_list, url, conf)
    assert child_list == [
        {
            "resources": [
                {
                    "name": "pets",
                    "resources": [
                        {
                            "name": "{petId}",
                            "methods": [
                                {
                                    "function_name": "PetsHandler",
                                    "handler": "pets_handler",
                                    "index": "subfolder/path_parameters_handlers.py",
                                    "method": "GET",
                                    "reference": "subfolder.path_parameters_handlers.pets_handler",
                                    "status_code": "200",
                                }
                            ],
                        }
                    ],
                },
                {
                    "name": "context",
                    "methods": [
                        {
                            "function_name": "WithContext",
                            "handler": "with_context",
                            "index": "subfolder/additional_args_handlers.py",
                            "method": "GET",
                            "reference": "subfolder.additional_args_handlers.with_context",
                            "status_code": "200",
                        }
                    ],
                },
            ]
        }
    ]


def test_add_resource_missing_paths_bug():
    hold_dict = {
        "/data/{key}/vector": {
            "GET": {
                "function_name": "VectorDataRoutes",
                "status_code": "200",
                "index": "handlers/vector_api_handlers.py",
                "handler": "vector_data_routes",
                "reference": "handlers.vector_api_handlers.vector_data_routes",
            }
        },
        "/data/{key}/vector/geojson": {
            "GET": {
                "function_name": "QueryVectorData",
                "status_code": "200",
                "index": "handlers/vector_api_handlers.py",
                "handler": "query_vector_data",
                "reference": "handlers.vector_api_handlers.query_vector_data",
            }
        },
    }

    resource = []
    for url, conf in sorted(hold_dict.items()):
        add_resource_v2(resource, url, conf)

    response = [
        {
            "resources": [
                {
                    "name": "data",
                    "resources": [
                        {
                            "name": "{key}",
                            "resources": [
                                {
                                    "methods": [
                                        {
                                            "function_name": "VectorDataRoutes",
                                            "handler": "vector_data_routes",
                                            "index": "handlers/vector_api_handlers.py",
                                            "method": "GET",
                                            "reference": "handlers.vector_api_handlers.vector_data_routes",
                                            "status_code": "200",
                                        }
                                    ],
                                    "name": "vector",
                                    "resources": [
                                        {
                                            "methods": [
                                                {
                                                    "function_name": "QueryVectorData",
                                                    "handler": "query_vector_data",
                                                    "index": "handlers/vector_api_handlers.py",
                                                    "method": "GET",
                                                    "reference": "handlers.vector_api_handlers.query_vector_data",
                                                    "status_code": "200",
                                                }
                                            ],
                                            "name": "geojson",
                                        }
                                    ],
                                }
                            ],
                        }
                    ],
                }
            ]
        }
    ]

    assert resource == response



import json

from demo_app_handlers import create_handler  # type: ignore


def test_get_response(requests_client, base_url):
    """
    test that the message is returned to the body
    """
    response = requests_client.get(f"{base_url}/hello")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World"}


def test_post_response(requests_client, base_url):
    """
    test that the message is returned to the body
    """
    body = {"name": "Foo", "description": "An optional description", "price": 45.2, "tax": 3.5}
    response = requests_client.post(f"{base_url}/hello", json=body)
    assert response.status_code == 201, response.json()
    assert response.json() == body


def test_post_invalid_body(requests_client, base_url):
    """
    test that the message is returned to the body
    """
    body = {
        "description": "An optional description",
        "price": 45.2,
    }
    response = requests_client.post(f"{base_url}/hello", json=body)
    assert response.status_code == 422
    assert response.json() == {
        "detail": [{"loc": ["body", "name"], "msg": "field required", "type": "value_error.missing"}]
    }


def test_post_invalid_empty_json(requests_client, base_url):
    """
    test that the message is returned to the body
    """
    response = requests_client.post(f"{base_url}/hello", json="")
    assert response.status_code == 422
    assert response.json() == {
        "detail": [
            {"loc": ["body", "name"], "msg": "field required", "type": "value_error.missing"},
            {"loc": ["body", "price"], "msg": "field required", "type": "value_error.missing"},
        ]
    }


def test_post_invalid_json(requests_client, base_url):
    """
    test that the message is returned to the body
    """
    response = create_handler({"path": f"{base_url}/hello", "body": "{"}, None)
    assert response["statusCode"] == 400
    assert json.loads(response["body"]) == {"detail": "JSONDecodeError"}


def test_inv(mock_lambda_context):
    event = {
        "resource": "/hello",
        "path": "/hello",
        "httpMethod": "POST",
        "queryStringParameters": None,
        "body": '{"name": "Foo", "description": "An optional description", "price": 45.2, "tax": 3.5}',
        "isBase64Encoded": False,
    }

    response = create_handler(event, mock_lambda_context)

    assert response["statusCode"] == 201, response.get("body")


def test_list_response(requests_client, base_url):
    response = requests_client.get(f"{base_url}/list_response")
    assert response.status_code == 200
    assert response.json() == [{"item_name": 1}]


def test_list_response_model(requests_client, base_url):
    response = requests_client.get(f"{base_url}/list_response_model")
    assert response.status_code == 200
    assert response.json() == [{"item_name": "secret", "item_value": None}]


def test_error_much_handler(requests_client, base_url):
    response = requests_client.get(f"{base_url}/error_much")
    assert response.status_code == 502
    assert response.json() == {"message": "Internal server error"}


def test_error_much_handler_post(requests_client, base_url):
    response = requests_client.post(f"{base_url}/error_much")
    assert response.status_code == 502
    assert response.json() == {"message": "Internal server error"}



#!/usr/bin/env python3

if __name__ == "__main__":
    import aws_cdk as cdk
    from src.demo_app_stack import DemoAppStack

    app = cdk.App()
    DemoAppStack(app, "demo-app", env={"region": "eu-west-2"})

    app.synth()



import pytest
from handler_app import plh


def test_path_parameters_without_typehint(requests_client, base_url):
    response = requests_client.get(f"{base_url}/pets/1")
    assert response.status_code == 200
    assert response.json() == {"pet_id": "1"}


def test_path_parameters_with_typehint(requests_client, base_url):
    response = requests_client.get(f"{base_url}/items/2")
    assert response.status_code == 200
    assert response.json() == {"item_id": 2}


def test_path_parameters_with_enum_typehint(requests_client, base_url):
    response = requests_client.get(f"{base_url}/item_enum/dog")
    assert response.status_code == 200
    assert response.json() == {"item_id": "dog"}


def test_path_parameters_with_typehint_typeerror(requests_client, base_url):
    response = requests_client.get(f"{base_url}/item_enum/cat")
    assert response.status_code == 422
    assert response.json() == {
        "detail": [
            {
                "ctx": {"enum_values": ["dog"]},
                "loc": ["path", "item_id"],
                "msg": "value is not a valid enumeration member; permitted: 'dog'",
                "type": "type_error.enum",
            }
        ]
    }


def test_path_parameters_with_path_default():
    # Fix me should only error on run otherwise we block
    # all the other handlers
    with pytest.raises(Exception):

        @plh.get("/items/{item_id}")
        def handler_with_path_default(item_id=2):
            return {"item_id": item_id}



def test_context(requests_client, base_url):
    response = requests_client.get(f"{base_url}/context")
    assert response.status_code == 200, response.json()
    assert isinstance(response.json()["context"], int)



from awslambdaric.lambda_context import LambdaContext
from handler_app import plh

fake_items_db = [{"item_name": "Foo"}, {"item_name": "Bar"}, {"item_name": "Baz"}]


@plh.get("/context")
def with_context(lambda_context: LambdaContext):
    return {"context": lambda_context.get_remaining_time_in_millis()}



from collections import defaultdict
from typing import Any

from awslambdaric.lambda_context import LambdaContext

from pydantic_lambda_handler.middleware import BaseHook


class MockRequests(BaseHook):
    testing_stuff: dict = defaultdict(dict)
    _testing_url = None
    _method = None

    @classmethod
    def method_init(cls, **kwargs):
        url = kwargs["url"]
        cls._method = kwargs["method"]

        cls._testing_url = url.replace("{", "(?P<").replace("}", r">\w+)")
        if cls._testing_url not in cls.testing_stuff["paths"]:
            cls.testing_stuff["paths"][cls._testing_url] = {cls._method: {}}
        else:
            cls.testing_stuff["paths"][cls._testing_url][cls._method] = {}

    @classmethod
    def pre_path(cls, **kwargs) -> None:
        func = kwargs["func"]
        cls.testing_stuff["paths"][cls._testing_url][cls._method]["handler"]["function"] = func

    @classmethod
    def pre_func(cls, event, context) -> tuple[dict, LambdaContext]:
        return event, context

    @classmethod
    def post_func(cls, body) -> Any:
        return body

    @classmethod
    def post_create_response(cls, **kwargs):
        create_response = kwargs["create_response"]

        cls.testing_stuff["paths"][cls._testing_url][cls._method]["handler"] = {"decorated_function": create_response}
        return {}


def add_resource(child_dict: dict, url):
    part, found, remaining = url.partition("/")
    if part:
        if part in child_dict.get("resources", {}):
            return add_resource(child_dict["resources"][part], remaining)

        last_resource: dict[str, dict] = {}
        if "resources" not in child_dict:
            child_dict["resources"] = {part: last_resource}
        else:
            child_dict["resources"].update({part: last_resource})

        return add_resource(child_dict["resources"][part], remaining)
    return child_dict


def to_camel_case(text):
    s = text.replace("-", " ").replace("_", " ")
    s = s.split()
    if len(text) == 0:
        return text.capitalize()
    return "".join(i.capitalize() for i in s)



from typing import Optional

from pydantic import BaseModel

from pydantic_lambda_handler.main import PydanticLambdaHandler


def test_gen_event_model():
    """should work with an empty event"""

    class EventModel(BaseModel):
        path: dict[str, str]
        query: dict[str, str]
        body: Optional[dict[str, str]]

    PydanticLambdaHandler.parse_event_to_model({}, EventModel)


def test_base_url():
    url = PydanticLambdaHandler._generate_base_url(
        {
            "path": f"/resource/123/data",
            "requestContext": {
                "path": "/StageName/resource/123/data",
                "domainName": "test.example.com",
            },
        }
    )

    assert url == "https://test.example.com/StageName"



def test_no_query(requests_client, base_url):
    response = requests_client.get(f"{base_url}/query")
    assert response.status_code == 200, response.json()
    assert response.json() == [{"item_name": "Foo"}, {"item_name": "Bar"}, {"item_name": "Baz"}]


def test_query(requests_client, base_url):
    response = requests_client.get(f"{base_url}/query", params={"skip": 1})
    assert response.status_code == 200, response.json()
    assert response.json() == [{"item_name": "Bar"}, {"item_name": "Baz"}]


def test_missing_query(requests_client, base_url):
    response = requests_client.get(f"{base_url}/query_required")
    assert response.status_code == 422, response.json()
    assert response.json() == {
        "detail": [{"loc": ["query", "secret"], "msg": "field required", "type": "value_error.missing"}]
    }


def test_query_param(requests_client, base_url):
    response = requests_client.get(f"{base_url}/query_param", params={"meat": "solid"})
    assert response.status_code == 200, response.json()
    assert response.json() == {"sausages": "solid"}


def test_query_multivalue_param(requests_client, base_url):
    response = requests_client.get(f"{base_url}/query_multivalue_param", params={"sausages": [5, 6, 7, 8]})
    assert response.status_code == 200, response.json()
    assert response.json() == {"sausages": [5, 6, 7, 8]}


def test_query_param_float(requests_client, base_url):
    response = requests_client.get(f"{base_url}/query_float", params={"item_name": "4"})
    assert response.status_code == 200, response.json()
    assert response.json() == {"item_name": 4.0}



import json
import subprocess
from pathlib import Path

from invoke import task

from pydantic_lambda_handler.gen_open_api_inspect import gen_open_api_inspect

root = Path(__name__).absolute().parents[1]
print(root)
demo_app_dir = root.joinpath("demo_app").absolute()
assert demo_app_dir.exists(), demo_app_dir


@task
def build_and_deploy(c):
    subprocess.run("cdk bootstrap", check=True, shell=True, cwd=demo_app_dir)
    deploy(c)


@task
def deploy(c):
    subprocess.run("cdk deploy --require-approval never", check=True, shell=True, cwd=demo_app_dir)


@task(help={"app_dir": "The app directory", "output_file_path": "The output directory"})
def generate_open_api_spec(c, app_dir, output_file_path):
    """
    Generate an open api spec and write to file
    """
    app_dir_path = Path(app_dir)
    app_dir_path.exists()

    schema, *_ = gen_open_api_inspect(app_dir_path)

    output_file_path = Path(output_file_path)
    output_file_path.parent.exists()

    schema = json.loads(schema)

    with output_file_path.open("w") as f:
        json.dump(schema, f, indent=4)



"""
The main class which you import and use as a decorator.
"""
import datetime
import functools
import json
import logging
import re
import sys
import traceback
from collections.abc import Iterable
from decimal import Decimal
from http import HTTPStatus
from inspect import isclass, signature
from typing import Any, Optional, Union

from awslambdaric.lambda_context import LambdaContext
from orjson import JSONDecodeError, loads
from pydantic import BaseModel, ValidationError, create_model

from pydantic_lambda_handler.middleware import BaseHook
from pydantic_lambda_handler.models import BaseOutput
from pydantic_lambda_handler.params import Header, Param, Query

JSON_DATA_TYPES = int, str, float, Decimal, bool, datetime.datetime, datetime.date


class PydanticLambdaHandler:
    """
    The decorator handle.
    """

    _hooks: list[type[BaseHook]] = []

    def __init__(
        self,
        *,
        title="PydanticLambdaHandler",
        version="0.0.0",
        hooks: Optional[Iterable[type[BaseHook]]] = None,
        logger=None,
    ):
        self.title = title
        self.version = version
        if hooks:
            PydanticLambdaHandler._hooks.extend(hooks)
        self.logger = logger or logging.getLogger(__name__)

    @classmethod
    def add_hook(cls, hook: type[BaseHook]):
        cls._hooks.append(hook)

    def get(
        self,
        url,
        *,
        status_code: Union[HTTPStatus, int] = HTTPStatus.OK,
        operation_id: Optional[str] = None,
        description: str = "Successful Response",
        function_name=None,
        response_model=None,
        logger=None,
        errors: Optional[list[tuple[Union[HTTPStatus, int], Any]]] = None,
    ):
        """Expect request with a GET method.

        :param url:
        :param status_code:
        :return:
        """
        method = "get"
        if logger:
            self.logger = logger
        return self.run_method(
            method, url, status_code, operation_id, description, function_name, response_model, errors
        )

    def post(
        self,
        url,
        *,
        status_code: Union[HTTPStatus, int] = HTTPStatus.CREATED,
        operation_id: Optional[str] = None,
        description: str = "Successful Response",
        function_name=None,
        response_model=None,
        logger=None,
        errors=None,
    ):
        """Expect request with a POST method.

        :param url:
        :param status_code:
        :return:
        """
        method = "post"
        if logger:
            self.logger = logger
        return self.run_method(
            method,
            url,
            status_code,
            operation_id,
            description,
            function_name,
            response_model,
            errors,
        )

    def run_method(  # noqa: C901 too complex
        self,
        method,
        url,
        status_code,
        operation_id,
        description,
        function_name,
        response_model,
        errors,
    ):
        for hook in self._hooks:
            hook.method_init(**locals())

        def create_response(func):
            for hook in self._hooks:
                hook.pre_path(**locals())

            sig = signature(func)

            if sig.parameters:
                EventModel = self.generate_event_model(url, sig)

            @functools.wraps(func)
            def wrapper_decorator(event, context: LambdaContext):
                try:
                    self.base_url = self._generate_base_url(event)

                    for hook in self._hooks:
                        event, context = hook.pre_func(event, context)

                    self.logger.debug(f"{event=}")
                    self.logger.debug(f"{context=}")

                    sig = signature(func)

                    if sig.parameters:
                        try:
                            event_model = self.parse_event_to_model(event, EventModel)
                        except ValidationError as e:
                            response = BaseOutput(
                                body=json.dumps({"detail": json.loads(e.json())}),
                                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                                is_base_64_encoded=False,
                            )
                            return loads(response.json())
                        except JSONDecodeError:
                            response = BaseOutput(
                                body=json.dumps({"detail": "JSONDecodeError"}),
                                status_code=HTTPStatus.BAD_REQUEST,
                                is_base_64_encoded=False,
                            )
                            return loads(response.json())

                        # Do something before
                        func_kwargs = {
                            **event_model.path.dict(),
                            **event_model.query.dict(),
                            **event_model.multiquery.dict(),
                            **event_model.headers.dict(),
                        }
                        if hasattr(event_model, "body"):
                            func_kwargs.update(**{event_model.body._alias: event_model.body})

                        if context_name := next(
                            (
                                i.name
                                for i in iter(sig.parameters.values())
                                if isinstance(i.annotation, type) and issubclass(i.annotation, LambdaContext)
                            ),
                            None,
                        ):
                            func_kwargs[context_name] = context
                    else:
                        func_kwargs = {}

                    try:
                        body = func(**func_kwargs)
                    except Exception as e:
                        if errors:
                            for e_status_code, exceptions in errors:
                                if isinstance(e, exceptions):
                                    if hasattr(e, "json"):
                                        body = json.dumps({"detail": json.loads(e.json())})
                                    else:
                                        body = json.dumps(
                                            {
                                                "detail": [
                                                    {
                                                        "msg": getattr(e, "msg", str(e)),
                                                        "type": type(e).__name__,
                                                    }
                                                ]
                                            }
                                        )
                                    response = BaseOutput(
                                        body=body, status_code=e_status_code, is_base_64_encoded=False
                                    )
                                    return loads(response.json())
                        raise

                    for hook in reversed(self._hooks):
                        body = hook.post_func(body)

                    if response_model:
                        body = response_model.parse_obj(body)

                    if hasattr(body, "json"):
                        base_output = BaseOutput(body=body.json(), status_code=status_code, is_base_64_encoded=False)
                    else:
                        base_output = BaseOutput(
                            body=json.dumps(body), status_code=status_code, is_base_64_encoded=False
                        )

                    for hook in self._hooks:
                        body = hook.pre_return(base_output)

                    response = loads(base_output.json())
                except Exception as error:
                    traceback.print_exc(file=sys.stdout)
                    self.logger.error(f"{type(error).__name__}: {error}")
                    raise
                else:
                    self.logger.debug(f"{context=}")
                    return response

            return wrapper_decorator

        for hook in self._hooks:
            hook.post_create_response(**locals())
        return create_response

    @staticmethod
    def generate_event_model(url, sig):
        path_model_dict = {}
        query_model_dict = {}
        multiquery_model_dict = {}

        body_default = None
        body_model = None

        headers = {}
        path_parameters_list = list(re.findall(r"\{(.*?)\}", url))
        path_parameters = set(path_parameters_list)
        additional_kwargs = {}
        if len(path_parameters_list) != len(path_parameters):
            raise ValueError(f"re-declared path variable: {url}")

        for param, param_info in sig.parameters.items():
            if isinstance(param_info.default, Header):
                headers[param] = param_info.annotation, param_info.default
                continue
            elif isinstance(param_info.default, Query):
                query_model_dict[param] = param_info.annotation, param_info.default
                continue
            elif param in path_parameters:
                if param_info.annotation == param_info.empty:
                    annotations = str, ...
                else:
                    annotations = param_info.annotation, ...
            else:
                default = ... if param_info.default == param_info.empty else param_info.default
                if param_info.annotation == param_info.empty:
                    annotations = str, default
                else:
                    annotations = param_info.annotation, default

            if param in path_parameters:
                if param_info.default != param_info.empty:
                    raise ValueError("Should not set default for path parameters")
                path_model_dict[param] = annotations
            else:
                model, body_default = annotations
                if isinstance(model, Param):
                    raise NotImplementedError
                elif isclass(param_info.annotation) and issubclass(model, BaseModel):
                    if body_model:
                        raise ValueError("Can only use one Pydantic model for body only")
                    body_model = model
                    body_model._alias = param
                elif isclass(param_info.annotation) and issubclass(model, LambdaContext):
                    additional_kwargs[param] = annotations
                elif isclass(annotations[0]) and issubclass(annotations[0], (JSON_DATA_TYPES)):
                    query_model_dict[param] = annotations
                elif annotations[0].__args__[0].__name__ == "list":
                    multiquery_model_dict[param] = annotations
                else:
                    query_model_dict[param] = annotations

                model, body_default = annotations
        if path_parameters != set(path_model_dict.keys()):
            raise ValueError("Missing path parameters")

        PathModel = create_model("PathModel", **path_model_dict)
        QueryModel = create_model("QueryModel", **query_model_dict)
        MultiValueQueryModel = create_model("MultiValueQueryModel", **multiquery_model_dict)
        HeaderModel = create_model("HeaderModel", **headers)
        event_models = {
            "path": (PathModel, {}),
            "query": (QueryModel, {}),
            "multiquery": (MultiValueQueryModel, {}),
            "headers": (HeaderModel, {}),
        }
        if body_model:
            event_models["body"] = (body_model, body_default)

        return create_model("EventModel", **event_models)

    @staticmethod
    def parse_event_to_model(event, EventModel):
        path_parameters = event.get("pathParameters", {}) or {}
        query_parameters = event.get("queryStringParameters", {}) or {}
        multiquery_parameters = event.get("multiValueQueryStringParameters", {}) or {}
        headers = event.get("headers", {}) or {}
        body = loads(event["body"]) if event.get("body") else None
        return EventModel(
            path=path_parameters, query=query_parameters, multiquery=multiquery_parameters, body=body, headers=headers
        )

    @staticmethod
    def _generate_base_url(event):
        """Returns the base url, including stage name"""
        path_prefix = event.get("requestContext", {}).get("path", "")[: -len(event.get("path", ""))].rstrip("/")
        domain = event.get("requestContext", {}).get("domainName", "")
        return f"https://{domain}{path_prefix}"



import subprocess
from pathlib import Path


def test_gen_cdk_conf():
    cwd = Path(__file__).parents[1].joinpath("demo_app")
    completed_process = subprocess.run(["gen-cdk-conf", "src", "cdk_conf.json"], cwd=cwd)
    assert completed_process.returncode == 0


def test_open_api_spec():
    cwd = Path(__file__).parents[1].joinpath("demo_app")
    completed_process = subprocess.run(["gen-open-api-spec", "src", "open_api_spec.json"], cwd=cwd)
    assert completed_process.returncode == 0



def test_with_headers(requests_client, base_url):
    headers = {"user_agent": "user_123"}
    response = requests_client.get(f"{base_url}/with_headers", headers=headers)
    assert response.status_code == 200, response.json()
    assert response.json() == {"user_agent": "user_123"}


def test_with_headers_default(requests_client, base_url):
    response = requests_client.get(f"{base_url}/with_headers")
    assert response.status_code == 200, response.json()
    assert response.json() == {"user_agent": None}


def test_with_header_alias(requests_client, base_url):
    headers = {"UserId": "user_123"}
    response = requests_client.get(f"{base_url}/with_headers_alias", headers=headers)
    assert response.status_code == 200, response.json()
    assert response.json() == {"user_agent": "user_123"}



import logging

from pydantic_lambda_handler.main import PydanticLambdaHandler

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)

plh = PydanticLambdaHandler(title="PydanticLambdaHandler", logger=logger)



"""
location of base models
"""
from http import HTTPStatus
from typing import Union

from pydantic import BaseModel, Field


class BaseOutput(BaseModel):
    """
    https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html#api-gateway-simple-proxy-for-lambda-output-format

    {
        "isBase64Encoded": true | false,
        "statusCode": httpStatusCode,
        "headers": {"headerName": "headerValue", ...},
        "multiValueHeaders": {
            "headerName": ["headerValue", "headerValue2", ...],
            ...
        },
        "body": "..."
    }
    """

    isBase64Encoded: bool = Field(False, alias="is_base_64_encoded")
    statusCode: Union[HTTPStatus, int] = Field(..., alias="status_code")
    headers: dict[str, str] = Field(default_factory=dict, description='{"headerName": "headerValue", ...}')
    multiValueHeaders: dict[str, Union[str, list[str]]] = Field(
        default_factory=dict,
        description='{"headerName": ["headerValue", "headerValue2", ...], ...}',
        alias="multi_value_headers",
    )
    body: Union[dict, list, str, int]



import inspect
import json
import re
from http import client
from inspect import signature
from typing import Any

from awslambdaric.lambda_context import LambdaContext
from openapi_schema_pydantic.v3.v3_0_3 import (
    Components,
    Info,
    OpenAPI,
    Operation,
    PathItem,
    Response,
    Schema,
)
from openapi_schema_pydantic.v3.v3_0_3.util import (
    PydanticSchema,
    construct_open_api_with_schema_class,
)
from pydantic import BaseModel, create_model

from pydantic_lambda_handler.main import PydanticLambdaHandler
from pydantic_lambda_handler.middleware import BaseHook
from pydantic_lambda_handler.params import Header, Query


class APIGenerationHook(BaseHook):
    """Gen open api"""

    title: str = "Pydantic Lambda Handler"
    version: str = "0.0.0"
    paths: dict[str, PathItem] = {}
    method = None
    schemas: dict[str, Schema] = {}

    @classmethod
    def method_init(cls, **kwargs):
        app: PydanticLambdaHandler = kwargs["self"]
        status_code = kwargs["status_code"]
        open_api_status_code = str(int(status_code))
        cls.method = kwargs["method"]
        response_model = kwargs["response_model"]

        APIGenerationHook.title = app.title
        APIGenerationHook.version = app.version

        if response_model:
            schema = {"schema": PydanticSchema(schema_class=response_model)}
        else:
            schema = {}

        url = kwargs["url"]
        responses = {
            open_api_status_code: Response(
                description=kwargs["description"],
                content={"application/json": schema},
            )
        }
        if kwargs.get("errors"):
            for e_status_code, errors in kwargs["errors"]:
                if inspect.isclass(errors) and issubclass(errors, Exception):
                    responses[int(e_status_code)] = Response(
                        description=getattr(errors, "description", None) or inspect.getdoc(errors),
                        content={"application/json": {}},
                    )
                else:
                    responses[int(e_status_code)] = Response(
                        description=client.responses[int(e_status_code)],
                        content={"application/json": {}},
                    )
        if url in cls.paths:
            setattr(
                cls.paths[url],
                cls.method,
                Operation(responses=responses),
            )

        else:
            cls.paths[url] = PathItem(
                **{cls.method: Operation(responses=responses)},
            )

        if kwargs["operation_id"]:
            getattr(cls.paths[url], cls.method).operationId = kwargs["operation_id"]

    @classmethod
    def pre_path(cls, **kwargs) -> None:  # noqa: C901 too complex
        sig = signature(kwargs["func"])

        if sig.parameters:
            url = kwargs["url"]

            path_model_dict = {}
            query_model_dict = {}

            body_model = None

            path_parameters_list = list(re.findall(r"\{(.*?)\}", url))
            path_parameters = set(path_parameters_list)

            headers = {}

            if len(path_parameters_list) != len(path_parameters):
                raise ValueError(f"re-declared path variable: {url}")

            for param, param_info in sig.parameters.items():
                if isinstance(param_info.default, Header):
                    headers[param] = param_info.annotation, param_info.default
                    continue
                elif isinstance(param_info.default, Query):
                    query_model_dict[param] = param_info.annotation, param_info.default
                    continue

                if inspect.isclass(param_info.annotation) and issubclass(param_info.annotation, LambdaContext):
                    continue

                if param in path_parameters:
                    if param_info.annotation == param_info.empty:
                        annotations = str, ...
                    else:
                        annotations = param_info.annotation, ...
                else:
                    default = ... if param_info.default == param_info.empty else param_info.default
                    if param_info.annotation == param_info.empty:
                        annotations = str, default
                    else:
                        annotations = param_info.annotation, default

                if param in path_parameters:
                    if param_info.default != param_info.empty:
                        raise ValueError("Should not set default for path parameters")
                    path_model_dict[param] = annotations
                else:
                    model, body_default = annotations
                    if inspect.isclass(param_info.annotation) and issubclass(model, BaseModel):
                        if body_model:
                            raise ValueError("Can only use one Pydantic model for body only")
                        body_model: BaseModel = model  # type: ignore
                        body_model._alias = param  # type: ignore
                    else:
                        query_model_dict[param] = annotations  # type: ignore

            if path_parameters != set(path_model_dict.keys()):
                raise ValueError("Missing path parameters")

            APIPathModel = create_model("APIPathModel", **path_model_dict, **query_model_dict, **headers)  # type: ignore

            path_schema_initial = APIPathModel.schema()
            properties = []
            for name, property_info in path_schema_initial.get("properties", {}).items():
                if "$ref" in property_info:
                    property_info["$ref"] = property_info["$ref"].replace("#/definitions/", "#/components/schemas/")
                    for key, value in path_schema_initial.get("definitions", {}).items():
                        cls.schemas[key] = Schema.parse_obj(value)

                # FIXME: This is ugly, please refactor
                alias_header = next((h for _, h in headers.values() if h.alias == name), None)
                if name in headers or alias_header:
                    if (name in headers and headers[name][1].include_in_schema is False) or (
                        alias_header and alias_header.include_in_schema is False
                    ):
                        continue

                    p = {"name": name, "in": "header", "schema": property_info}
                    if name in path_schema_initial.get("required", ()):
                        p["required"] = True
                    properties.append(p)
                    continue

                in_ = "path" if name in path_parameters else "query"
                p = {"name": name, "in": in_, "schema": property_info}
                if name in path_schema_initial.get("required", ()):
                    p["required"] = True

                properties.append(p)

            getattr(cls.paths[url], cls.method).parameters = properties  # type: ignore
            if body_model:
                getattr(cls.paths[url], cls.method).requestBody = {  # type: ignore
                    "content": {"application/json": {"schema": PydanticSchema(schema_class=body_model)}}
                }

    @classmethod
    def pre_func(cls, event, context) -> tuple[dict, LambdaContext]:
        return event, context

    @classmethod
    def post_func(cls, body) -> Any:
        return body

    @classmethod
    def generate(cls):
        open_api = OpenAPI(
            info=Info(title=cls.title, version=cls.version), paths=cls.paths, components=Components(schemas=cls.schemas)
        )
        open_api = construct_open_api_with_schema_class(open_api)

        # sort keys
        open_api_dict = json.loads(open_api.json(by_alias=True, exclude_none=True))

        open_api_dict["paths"] = dict(sorted(open_api_dict["paths"].items()))
        return json.dumps(open_api_dict, indent=2)



import logging
from typing import Optional

from handler_app import plh
from pydantic import BaseModel

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)


class Item(BaseModel):
    name: str
    description: Optional[str] = None
    price: float
    tax: Optional[float] = None


@plh.post("/hello", logger=logger)
def create_handler(item: Item):
    return item


@plh.get("/hello")
def hello_handler():
    from pydantic_core import SchemaValidator, ValidationError

    v = SchemaValidator(
        {
            "type": "typed-dict",
            "fields": {
                "name": {
                    "schema": {
                        "type": "str",
                    },
                },
                "age": {
                    "schema": {
                        "type": "int",
                        "ge": 18,
                    },
                },
                "is_developer": {
                    "schema": {
                        "type": "bool",
                    },
                },
            },
        }
    )

    r1 = v.validate_python({"name": "Samuel", "age": 35, "is_developer": True})
    assert r1 == {"name": "Samuel", "age": 35, "is_developer": True}

    # pydantic-core can also validate JSON directly
    r2 = v.validate_json('{"name": "Samuel", "age": 35, "is_developer": true}')
    assert r1 == r2

    try:
        v.validate_python({"name": "Samuel", "age": 11})
    except ValidationError:
        """
        1 validation error for model
        age
          Input should be greater than or equal to 18
          [kind=greater_than_equal, context={ge: 18}, input_value=11, input_type=int]
        """
    return {"message": "Hello World"}


@plh.post("/")
def index_handler():
    return {"message": "Hello Index"}


@plh.get("/error_much", logger=logger)
def error_much_handler():
    raise ValueError("it's broken")


@plh.post("/error_much", logger=logger)
def error_much_handler_post():
    raise ValueError("it's broken")



from typing import Union

from handler_app import plh

from pydantic_lambda_handler.params import Header


@plh.get("/with_headers")
def with_headers(user_agent: Union[str, None] = Header(default=None)):
    return {"user_agent": user_agent}


@plh.get("/with_headers_not_in_schema")
def with_headers_not_in_schema(user_agent: Union[str, None] = Header(default=None, include_in_schema=False)):
    return {"user_agent": user_agent}


@plh.get("/with_headers_alias")
def with_headers_alias(user_agent: Union[str, None] = Header(default=None, alias="UserId")):
    return {"user_agent": user_agent}



import json
from http import client

from openapi_spec_validator import validate_spec

from pydantic_lambda_handler.hooks.open_api_gen_hook import APIGenerationHook
from pydantic_lambda_handler.main import PydanticLambdaHandler
from pydantic_lambda_handler.params import Header


def test_generate_open_api_version(schema):
    assert schema["openapi"] == "3.0.3"


def test_valid_open_api_spc(schema):
    validate_spec(schema)


def test_valid_open_api_spc_missing_key():
    app = PydanticLambdaHandler(title="PydanticLambdaHandler")
    app.add_hook(APIGenerationHook)

    @app.get("/sources/{key}/data")
    def data_links(key: str):
        return {"key": key}

    schema = next(h for h in app._hooks if issubclass(h, APIGenerationHook)).generate()  # type: ignore
    schema = json.loads(schema)
    validate_spec(schema)


def test_valid_open_api_spc_missing_key_with_header():
    app = PydanticLambdaHandler(title="PydanticLambdaHandler")
    app.add_hook(APIGenerationHook)
    header_host = Header("", alias="Host", include_in_schema=False)

    @app.get("/sources/{key}/data")
    def data_links(key: str, host: str = header_host):
        return {"key": key}

    schema = next(h for h in app._hooks if issubclass(h, APIGenerationHook)).generate()  # type: ignore
    schema = json.loads(schema)
    validate_spec(schema)


def test_generate_open_api_info(schema):
    assert schema["info"] == {"title": "PydanticLambdaHandler", "version": "0.0.0"}


def test_generate_open_api_info_path_get(schema):
    item_path = schema["paths"]["/hello"]["get"]["responses"]["200"]["content"]
    assert item_path == {"application/json": {}}


def test_generate_open_api_info_path_post(schema):
    item_path = schema["paths"]["/hello"]["post"]["responses"]["201"]["content"]
    assert item_path == {"application/json": {}}


def test_generate_open_api_list_response_model(schema):
    item_path = schema["paths"]["/list_response_model"]["get"]["responses"]["200"]["content"]
    assert item_path == {"application/json": {"schema": {"$ref": "#/components/schemas/ListFunModel"}}}
    assert schema["components"]["schemas"]["ListFunModel"] == {
        "items": {"$ref": "#/components/schemas/FunModel"},
        "title": "ListFunModel",
        "type": "array",
    }


def test_query_body(schema):
    request_schema = schema["paths"]["/hello"]["post"]["requestBody"]["content"]["application/json"]["schema"]
    assert request_schema == {"$ref": "#/components/schemas/Item"}
    assert schema["components"]["schemas"]["Item"] == {
        "properties": {
            "description": {"title": "Description", "type": "string"},
            "name": {"title": "Name", "type": "string"},
            "price": {"title": "Price", "type": "number"},
            "tax": {"title": "Tax", "type": "number"},
        },
        "required": ["name", "price"],
        "title": "Item",
        "type": "object",
    }


def test_generate_open_api_status_code_int(schema):
    """Can accept an in or an Enum status code"""
    assert "418" in schema["paths"]["/teapot"]["get"]["responses"]


def test_generate_open_api_path(schema):
    assert "/pets/{petId}" in schema["paths"]
    assert schema["paths"]["/pets/{petId}"]["get"].get("parameters") == [
        {"name": "petId", "in": "path", "required": True, "schema": {"title": "Petid", "type": "string"}}
    ]


def test_generate_open_operation_id(schema):
    assert schema["paths"]["/pets/{petId}"]["get"].get("operationId") == "Create Pet"


def test_query_options(schema):
    assert "/query" in schema["paths"]
    assert schema["paths"]["/query"]["get"].get("parameters") == [
        {"in": "query", "name": "skip", "schema": {"default": 0, "title": "Skip", "type": "integer"}},
        {"in": "query", "name": "limit", "schema": {"default": 10, "title": "Limit", "type": "integer"}},
    ]


def test_response_body(schema):
    assert "/response_model" in schema["paths"]
    response_schema = schema["paths"]["/response_model"]["get"]["responses"]["200"]["content"]["application/json"][
        "schema"
    ]
    assert response_schema == {"$ref": "#/components/schemas/FunModel"}
    assert schema["components"]["schemas"]["FunModel"] == {
        "title": "FunModel",
        "required": ["item_name"],
        "type": "object",
        "properties": {
            "item_name": {"title": "Item Name", "type": "string"},
            "item_value": {"title": "Item Value", "type": "integer"},
        },
    }


def test_header_options(schema):
    assert "/with_headers" in schema["paths"]
    assert schema["paths"]["/with_headers"]["get"].get("parameters") == [
        {"in": "header", "name": "user_agent", "schema": {"title": "User Agent", "type": "string"}}
    ]


def test_header_options_not_in_schema(schema):
    assert "/with_headers_not_in_schema" in schema["paths"]
    assert schema["paths"]["/with_headers_not_in_schema"]["get"].get("parameters") == []


def test_header_options_uses_alias(schema):
    assert "/with_headers_alias" in schema["paths"]
    assert schema["paths"]["/with_headers_alias"]["get"].get("parameters") == [
        {"in": "header", "name": "UserId", "schema": {"title": "Userid", "type": "string"}}
    ]


def test_errors(schema):
    assert "/error" in schema["paths"]
    assert schema["paths"]["/error"]["get"]["responses"]["418"] == {
        "content": {"application/json": {}},
        "description": "Inappropriate argument value (of correct type).",
    }


def test_multiple_errors_description(schema):
    assert "/multiple_errors" in schema["paths"]
    description = schema["paths"]["/multiple_errors"]["get"]["responses"]["422"]["description"]
    assert description == client.responses[422]


def test_query_union(schema):
    assert "/query_union" in schema["paths"]
    parameters = schema["paths"]["/query_union"]["get"]["parameters"]
    assert parameters == [
        {
            "name": "param",
            "in": "query",
            "schema": {"title": "Param", "anyOf": [{"type": "number"}, {"type": "string", "format": "date-time"}]},
        }
    ]



def test_error(requests_client, base_url):
    response = requests_client.get(f"{base_url}/error")
    assert response.status_code == 418, response.json()
    assert response.json() == {"detail": [{"msg": "nope", "type": "ValueError"}]}
